/***
  This file is part of avahi.

  avahi is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2.1 of the
  License, or (at your option) any later version.

  avahi is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
  Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with avahi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.
***/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_BONJOUR
#include <netinet/in.h> /* for ntohs */
#endif

#include <avahi-common/domain.h>
#include <avahi-common/timeval.h>
#include <avahi-common/malloc.h>
#include <avahi-common/error.h>

#include "browse.h"
#include "log.h"

#define TIMEOUT_MSEC 5000

struct AvahiSServiceResolver {
    AvahiServer *server;
    char *service_name;
    char *service_type;
    char *domain_name;
    AvahiProtocol address_protocol;

    AvahiIfIndex interface;
    AvahiProtocol protocol;

    AvahiSRecordBrowser *record_browser_srv;
    AvahiSRecordBrowser *record_browser_txt;
    AvahiSRecordBrowser *record_browser_a;
    AvahiSRecordBrowser *record_browser_aaaa;

    AvahiRecord *srv_record, *txt_record, *address_record;
    AvahiLookupResultFlags srv_flags, txt_flags, address_flags;

    AvahiSServiceResolverCallback callback;
    void* userdata;
    AvahiLookupFlags user_flags;

    AvahiTimeEvent *time_event;

#ifdef HAVE_BONJOUR
    uint16_t port;
    AvahiTimeEvent *defer_time_event;
    char *host;
    AvahiWatch *watch;
    AvahiWatch *watch_a;
    AvahiWatch *watch_a6;
    DNSServiceRef client;
    DNSServiceRef client_a;
    DNSServiceRef client_a6;
#endif

    AVAHI_LLIST_FIELDS(AvahiSServiceResolver, resolver);
};


#ifdef HAVE_BONJOUR
static void finish(AvahiSServiceResolver *r, AvahiResolverEvent event ) {
    AvahiLookupResultFlags flags;

    assert(r);

    if (r->time_event) {
        avahi_time_event_free(r->time_event);
        r->time_event = NULL;
    }

    flags = 0;

    switch (event) {
        case AVAHI_RESOLVER_FAILURE:

            r->callback(r,
                r->interface,
                r->protocol,
                event,
                r->service_name,
                r->service_type,
                r->domain_name,
                NULL,
                NULL,
                0,
                NULL,
                flags,
                r->userdata);
            break;

        case AVAHI_RESOLVER_FOUND: {
            AvahiAddress a;

            if (r->address_record) {
                switch (r->address_record->key->type) {
                    case AVAHI_DNS_TYPE_A:
                        a.proto = AVAHI_PROTO_INET;
                        a.data.ipv4 = r->address_record->data.a.address;
                        break;

                    case AVAHI_DNS_TYPE_AAAA:
                        a.proto = AVAHI_PROTO_INET6;
                        a.data.ipv6 = r->address_record->data.aaaa.address;
                        break;

                    default:
                        assert(0);
                }
            }

            flags = 0;
            if (avahi_domain_equal(r->server->host_name_fqdn, r->host))
                flags |= AVAHI_LOOKUP_RESULT_LOCAL;
            r->callback(r,
                r->interface,
                r->protocol,
                event,
                r->service_name,
                r->service_type,
                r->domain_name,
                r->host,
                r->address_record ? &a : NULL,
                 r->port,
                r->txt_record ? r->txt_record->data.txt.string_list : NULL,
                flags,
                r->userdata);
            break;
        }
    }
}

static  void query_record(DNSServiceRef sdRef, const DNSServiceFlags flags, uint32_t ifIndex, DNSServiceErrorType errorCode,
        const char *fullname, uint16_t rrtype, uint16_t rrclass, uint16_t rdlen, const void *rdata, uint32_t ttl, void *context)
{
    AvahiSServiceResolver *r = context;
    const unsigned char *rd = rdata;
    AvahiAddress a;
    int changed = 0;

    if (r->interface > 0 && ifIndex != r->interface)
        return;
    if (r->interface <= 0)
        r->interface = ifIndex;

    switch (rrtype) {
        case kDNSServiceType_A:
            if (!r->address_record) {
                if (!(r->address_record = avahi_record_new_full(r->host, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_A, ttl))) {
                    avahi_server_set_errno(r->server, AVAHI_ERR_NO_MEMORY);
                    finish(r, AVAHI_RESOLVER_FAILURE);
                    return;
                }
                if (avahi_rdata_parse(r->address_record, rdata, rdlen) != 0) {
                    avahi_server_set_errno(r->server, AVAHI_ERR_INVALID_PACKET);
                    finish(r, AVAHI_RESOLVER_FAILURE);
                }
                changed = 1;
                if (r->protocol == AVAHI_PROTO_UNSPEC)
                    r->protocol = AVAHI_PROTO_INET;
            }
            break;

        case kDNSServiceType_A6:
            if (!r->address_record) {
                if (!(r->address_record = avahi_record_new_full(r->host, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_AAAA, ttl))) {
                    avahi_server_set_errno(r->server, AVAHI_ERR_NO_MEMORY);
                    finish(r, AVAHI_RESOLVER_FAILURE);
                    return;
                }
                if (avahi_rdata_parse(r->address_record, rdata, rdlen) != 0) {
                    avahi_server_set_errno(r->server, AVAHI_ERR_INVALID_PACKET);
                    finish(r, AVAHI_RESOLVER_FAILURE);
                }
                changed = 1;
                if (r->protocol == AVAHI_PROTO_UNSPEC)
                    r->protocol = AVAHI_PROTO_INET6;
            }
            break;
        default:
            abort();

    }

    if (changed &&
        (r->txt_record || (r->user_flags & AVAHI_LOOKUP_NO_TXT)) &&
        r->address_record || (r->user_flags & AVAHI_LOOKUP_NO_ADDRESS)) {
        finish(r, AVAHI_RESOLVER_FOUND);
   }
}

static void resolve_socket_event(AvahiWatch *w, int fd, AvahiWatchEvent events, void *userdata) {
    AvahiSServiceResolver *r = userdata;
    DNSServiceErrorType ret;
    DNSServiceRef client;

    assert(w);
    assert(fd >= 0);
    assert(events & AVAHI_WATCH_IN);

    if (fd == DNSServiceRefSockFD(r->client))
        client = r->client;
    else if (fd == DNSServiceRefSockFD(r->client_a))
        client = r->client_a;
    else if (fd == DNSServiceRefSockFD(r->client_a6))
        client = r->client_a6;
    else
        assert (0);

    ret = DNSServiceProcessResult(client);
    if (ret != kDNSServiceErr_NoError) {
        if (client == r->client) {
            if (r->watch) {
                r->server->poll_api->watch_free(r->watch);
                r->watch = NULL;
            }
            DNSServiceRefDeallocate(r->client);
            r->client = NULL;
        } else if (client == r->client_a) {
            if (r->watch_a) {
                r->server->poll_api->watch_free(r->watch_a);
                r->watch_a = NULL;
            }
            DNSServiceRefDeallocate(r->client_a);
            r->client_a = NULL;
        } else if (client == r->client_a6) {
            if (r->watch_a6) {
                r->server->poll_api->watch_free(r->watch_a6);
                r->watch_a6 = NULL;
            }
            DNSServiceRefDeallocate(r->client_a6);
            r->client_a6 = NULL;
        }
        avahi_server_set_errno(r->server, AVAHI_ERR_DISCONNECTED);
        finish(r, AVAHI_RESOLVER_FAILURE);
    }
}

static void resolve_reply(DNSServiceRef client, const DNSServiceFlags flags, uint32_t ifIndex, DNSServiceErrorType errorCode,
        const char *fullname, const char *host, uint16_t opaqueport, uint16_t txtlen, const char* txtrecord, void *context) {
    AvahiSServiceResolver *r = context;
    AvahiServer *s = r->server;
    DNSServiceErrorType ret;
    DNSServiceRef new_client;

    if (r->interface > 0 && ifIndex != r->interface)
        return;
    if (r->interface <= 0)
        r->interface = ifIndex;

    if (host) {
        if (!(r->host = avahi_strdup(host))) {
            avahi_server_set_errno(r->server, AVAHI_ERR_NO_MEMORY);
            finish(r, AVAHI_RESOLVER_FAILURE);
        }
    }
    if (!(r->user_flags & AVAHI_LOOKUP_NO_ADDRESS)) {
        if (!r->txt_record) {
            if (!(r->txt_record = avahi_record_new_full(r->host, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_TXT, AVAHI_DEFAULT_TTL))) {
                avahi_server_set_errno(r->server, AVAHI_ERR_NO_MEMORY);
                finish(r, AVAHI_RESOLVER_FAILURE);
                return;
            }
        }
        if (avahi_rdata_parse(r->txt_record, txtrecord, txtlen) != 0) {
            avahi_server_set_errno(r->server, AVAHI_ERR_INVALID_PACKET);
            finish(r, AVAHI_RESOLVER_FAILURE);
        }
    }
    r->port = ntohs(opaqueport);
    if (!(r->user_flags & AVAHI_LOOKUP_NO_ADDRESS)) {
        if (r->address_protocol == AVAHI_PROTO_INET || r->address_protocol == AVAHI_PROTO_UNSPEC) {
            if (r->client_a) {
                if (r->watch_a) {
                    r->server->poll_api->watch_free(r->watch_a);
                    r->watch_a = NULL;
                }
                DNSServiceRefDeallocate(r->client_a);
                r->client_a = NULL;
            }
            ret = DNSServiceQueryRecord(&r->client_a, 0, ifIndex, host, kDNSServiceType_A, kDNSServiceClass_IN, query_record, r);
            if (ret != kDNSServiceErr_NoError || !r->client_a) {
                finish(r, AVAHI_RESOLVER_FAILURE);
                return;
            }
            r->watch_a = s->poll_api->watch_new(s->poll_api, DNSServiceRefSockFD(r->client_a), AVAHI_WATCH_IN, resolve_socket_event, r);
        }

        if (r->address_protocol == AVAHI_PROTO_INET6 || r->address_protocol == AVAHI_PROTO_UNSPEC) {
            if (r->client_a6) {
                if (r->watch_a6) {
                    r->server->poll_api->watch_free(r->watch_a6);
                    r->watch_a6 = NULL;
                }
                DNSServiceRefDeallocate(r->client_a6);
                r->client_a6 = NULL;
            }
            ret = DNSServiceQueryRecord(&r->client_a6, 0, ifIndex, host, kDNSServiceType_A6, kDNSServiceClass_IN, query_record, r);
            if (ret != kDNSServiceErr_NoError || !r->client_a6) {
                finish(r, AVAHI_RESOLVER_FAILURE);
                return;
            }
            r->watch_a6 = s->poll_api->watch_new(s->poll_api, DNSServiceRefSockFD(r->client_a6), AVAHI_WATCH_IN, resolve_socket_event, r);
        }
    } else {
       finish(r, AVAHI_RESOLVER_FOUND);
    }
}

static void resolve_error_callback(AvahiTimeEvent *e, void *userdata) {
    AvahiSServiceResolver *r = userdata;

    if (r->defer_time_event) {
        avahi_time_event_free(r->defer_time_event);
        r->defer_time_event = NULL;
    }
    avahi_server_set_errno(r->server, AVAHI_ERR_FAILURE);
    finish(r, AVAHI_RESOLVER_FAILURE);
}

static void avahi_resolve_service_start(AvahiServer *s, AvahiSServiceResolver *r) {
    DNSServiceErrorType ret;

    ret = DNSServiceResolve(&r->client,
              0,
              r->interface == AVAHI_IF_UNSPEC ?
                  kDNSServiceInterfaceIndexAny :
                  r->interface,
              r->service_name,
              r->service_type,
              r->domain_name,
              resolve_reply,
              r);
    if (ret != kDNSServiceErr_NoError || !r->client) {
        r->defer_time_event = avahi_time_event_new(r->server->time_event_queue, NULL, resolve_error_callback, r);
    } else {
        r->watch = s->poll_api->watch_new(s->poll_api, DNSServiceRefSockFD(r->client), AVAHI_WATCH_IN, resolve_socket_event, r);
    }
}
#else

static void finish(AvahiSServiceResolver *r, AvahiResolverEvent event) {
    AvahiLookupResultFlags flags;

    assert(r);

    if (r->time_event) {
        avahi_time_event_free(r->time_event);
        r->time_event = NULL;
    }

    flags =
        r->txt_flags |
        r->srv_flags |
        r->address_flags;

    switch (event) {
        case AVAHI_RESOLVER_FAILURE:

            r->callback(
                r,
                r->interface,
                r->protocol,
                event,
                r->service_name,
                r->service_type,
                r->domain_name,
                NULL,
                NULL,
                0,
                NULL,
                flags,
                r->userdata);

            break;

        case AVAHI_RESOLVER_FOUND: {
            AvahiAddress a;

            assert(event == AVAHI_RESOLVER_FOUND);

            assert(r->srv_record);

            if (r->address_record) {
                switch (r->address_record->key->type) {
                    case AVAHI_DNS_TYPE_A:
                        a.proto = AVAHI_PROTO_INET;
                        a.data.ipv4 = r->address_record->data.a.address;
                        break;

                    case AVAHI_DNS_TYPE_AAAA:
                        a.proto = AVAHI_PROTO_INET6;
                        a.data.ipv6 = r->address_record->data.aaaa.address;
                        break;

                    default:
                        assert(0);
                }
            }

            r->callback(
                r,
                r->interface,
                r->protocol,
                event,
                r->service_name,
                r->service_type,
                r->domain_name,
                r->srv_record->data.srv.name,
                r->address_record ? &a : NULL,
                r->srv_record->data.srv.port,
                r->txt_record ? r->txt_record->data.txt.string_list : NULL,
                flags,
                r->userdata);

            break;
        }
    }
}
#endif

static void time_event_callback(AvahiTimeEvent *e, void *userdata) {
    AvahiSServiceResolver *r = userdata;

    assert(e);
    assert(r);

    avahi_server_set_errno(r->server, AVAHI_ERR_TIMEOUT);
    finish(r, AVAHI_RESOLVER_FAILURE);
}

static void start_timeout(AvahiSServiceResolver *r) {
    struct timeval tv;
    assert(r);

    if (r->time_event)
        return;

    avahi_elapse_time(&tv, TIMEOUT_MSEC, 0);

    r->time_event = avahi_time_event_new(r->server->time_event_queue, &tv, time_event_callback, r);
}

#ifndef HAVE_BONJOUR
static void record_browser_callback(
    AvahiSRecordBrowser*rr,
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    AvahiBrowserEvent event,
    AvahiRecord *record,
    AvahiLookupResultFlags flags,
    void* userdata) {

    AvahiSServiceResolver *r = userdata;

    assert(rr);
    assert(r);

    if (rr == r->record_browser_aaaa || rr == r->record_browser_a)
        r->address_flags = flags;
    else if (rr == r->record_browser_srv)
        r->srv_flags = flags;
    else if (rr == r->record_browser_txt)
        r->txt_flags = flags;

    switch (event) {

        case AVAHI_BROWSER_NEW: {
            int changed = 0;
            assert(record);

            if (r->interface > 0 && interface > 0 &&  interface != r->interface)
                return;

            if (r->protocol != AVAHI_PROTO_UNSPEC && protocol != AVAHI_PROTO_UNSPEC && protocol != r->protocol)
                return;

            if (r->interface <= 0)
                r->interface = interface;

            if (r->protocol == AVAHI_PROTO_UNSPEC)
                r->protocol = protocol;

            switch (record->key->type) {
                case AVAHI_DNS_TYPE_SRV:
                    if (!r->srv_record) {
                        r->srv_record = avahi_record_ref(record);
                        changed = 1;

                        if (r->record_browser_a) {
                            avahi_s_record_browser_free(r->record_browser_a);
                            r->record_browser_a = NULL;
                        }

                        if (r->record_browser_aaaa) {
                            avahi_s_record_browser_free(r->record_browser_aaaa);
                            r->record_browser_aaaa = NULL;
                        }

                        if (!(r->user_flags & AVAHI_LOOKUP_NO_ADDRESS)) {

                            if (r->address_protocol == AVAHI_PROTO_INET || r->address_protocol == AVAHI_PROTO_UNSPEC) {
                                AvahiKey *k = avahi_key_new(r->srv_record->data.srv.name, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_A);
                                r->record_browser_a = avahi_s_record_browser_prepare(r->server, r->interface, r->protocol, k, r->user_flags & ~(AVAHI_LOOKUP_NO_TXT|AVAHI_LOOKUP_NO_ADDRESS), record_browser_callback, r);
                                if(r->record_browser_a)
                                    avahi_s_record_browser_start_query(r->record_browser_a);
                                avahi_key_unref(k);
                            }

                            if (r->address_protocol == AVAHI_PROTO_INET6 || r->address_protocol == AVAHI_PROTO_UNSPEC) {
                                AvahiKey *k = avahi_key_new(r->srv_record->data.srv.name, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_AAAA);
                                r->record_browser_aaaa = avahi_s_record_browser_prepare(r->server, r->interface, r->protocol, k, r->user_flags & ~(AVAHI_LOOKUP_NO_TXT|AVAHI_LOOKUP_NO_ADDRESS), record_browser_callback, r);
                                if(r->record_browser_aaaa)
                                    avahi_s_record_browser_start_query(r->record_browser_aaaa);
                                avahi_key_unref(k);
                            }
                        }
                    }
                    break;

                case AVAHI_DNS_TYPE_TXT:

                    assert(!(r->user_flags & AVAHI_LOOKUP_NO_TXT));

                    if (!r->txt_record) {
                        r->txt_record = avahi_record_ref(record);
                        changed = 1;
                    }
                    break;

                case AVAHI_DNS_TYPE_A:
                case AVAHI_DNS_TYPE_AAAA:

                    assert(!(r->user_flags & AVAHI_LOOKUP_NO_ADDRESS));

                    if (!r->address_record) {
                        r->address_record = avahi_record_ref(record);
                        changed = 1;
                    }
                    break;

                default:
                    abort();
            }


            if (changed &&
                r->srv_record &&
                (r->txt_record || (r->user_flags & AVAHI_LOOKUP_NO_TXT)) &&
                (r->address_record || (r->user_flags & AVAHI_LOOKUP_NO_ADDRESS)))
                finish(r, AVAHI_RESOLVER_FOUND);

            break;

        }

        case AVAHI_BROWSER_REMOVE:

            assert(record);

            switch (record->key->type) {
                case AVAHI_DNS_TYPE_SRV:

                    if (r->srv_record && avahi_record_equal_no_ttl(record, r->srv_record)) {
                        avahi_record_unref(r->srv_record);
                        r->srv_record = NULL;

                        if (r->record_browser_a) {
                            avahi_s_record_browser_free(r->record_browser_a);
                            r->record_browser_a = NULL;
                        }

                        if (r->record_browser_aaaa) {
                            avahi_s_record_browser_free(r->record_browser_aaaa);
                            r->record_browser_aaaa = NULL;
                        }

                        /** Look for a replacement */
                        avahi_s_record_browser_restart(r->record_browser_srv);
                        start_timeout(r);
                    }

                    break;

                case AVAHI_DNS_TYPE_TXT:

                    assert(!(r->user_flags & AVAHI_LOOKUP_NO_TXT));

                    if (r->txt_record && avahi_record_equal_no_ttl(record, r->txt_record)) {
                        avahi_record_unref(r->txt_record);
                        r->txt_record = NULL;

                        /** Look for a replacement */
                        avahi_s_record_browser_restart(r->record_browser_txt);
                        start_timeout(r);
                    }
                    break;

                case AVAHI_DNS_TYPE_A:
                case AVAHI_DNS_TYPE_AAAA:

                    assert(!(r->user_flags & AVAHI_LOOKUP_NO_ADDRESS));

                    if (r->address_record && avahi_record_equal_no_ttl(record, r->address_record)) {
                        avahi_record_unref(r->address_record);
                        r->address_record = NULL;

                        /** Look for a replacement */
                        if (r->record_browser_aaaa)
                            avahi_s_record_browser_restart(r->record_browser_aaaa);
                        if (r->record_browser_a)
                            avahi_s_record_browser_restart(r->record_browser_a);
                        start_timeout(r);
                    }
                    break;

                default:
                    abort();
            }

            break;

        case AVAHI_BROWSER_CACHE_EXHAUSTED:
        case AVAHI_BROWSER_ALL_FOR_NOW:
            break;

        case AVAHI_BROWSER_FAILURE:

            if (rr == r->record_browser_a && r->record_browser_aaaa) {
                /* We were looking for both AAAA and A, and the other query is still living, so we'll not die */
                avahi_s_record_browser_free(r->record_browser_a);
                r->record_browser_a = NULL;
                break;
            }

            if (rr == r->record_browser_aaaa && r->record_browser_a) {
                /* We were looking for both AAAA and A, and the other query is still living, so we'll not die */
                avahi_s_record_browser_free(r->record_browser_aaaa);
                r->record_browser_aaaa = NULL;
                break;
            }

            /* Hmm, everything's lost, tell the user */

            if (r->record_browser_srv)
                avahi_s_record_browser_free(r->record_browser_srv);
            if (r->record_browser_txt)
                avahi_s_record_browser_free(r->record_browser_txt);
            if (r->record_browser_a)
                avahi_s_record_browser_free(r->record_browser_a);
            if (r->record_browser_aaaa)
                avahi_s_record_browser_free(r->record_browser_aaaa);

            r->record_browser_srv = r->record_browser_txt = r->record_browser_a = r->record_browser_aaaa = NULL;

            finish(r, AVAHI_RESOLVER_FAILURE);
            break;
    }
}
#endif

AvahiSServiceResolver *avahi_s_service_resolver_prepare(
    AvahiServer *server,
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    const char *name,
    const char *type,
    const char *domain,
    AvahiProtocol aprotocol,
    AvahiLookupFlags flags,
    AvahiSServiceResolverCallback callback,
    void* userdata) {

    AvahiSServiceResolver *r;
    AvahiKey *k;
    char n[AVAHI_DOMAIN_NAME_MAX];
    int ret;

    assert(server);
    assert(type);
    assert(callback);

    AVAHI_CHECK_VALIDITY_RETURN_NULL(server, AVAHI_IF_VALID(interface), AVAHI_ERR_INVALID_INTERFACE);
    AVAHI_CHECK_VALIDITY_RETURN_NULL(server, AVAHI_PROTO_VALID(protocol), AVAHI_ERR_INVALID_PROTOCOL);
    AVAHI_CHECK_VALIDITY_RETURN_NULL(server, AVAHI_PROTO_VALID(aprotocol), AVAHI_ERR_INVALID_PROTOCOL);
    AVAHI_CHECK_VALIDITY_RETURN_NULL(server, !domain || avahi_is_valid_domain_name(domain), AVAHI_ERR_INVALID_DOMAIN_NAME);
    AVAHI_CHECK_VALIDITY_RETURN_NULL(server, !name || avahi_is_valid_service_name(name), AVAHI_ERR_INVALID_SERVICE_NAME);
    AVAHI_CHECK_VALIDITY_RETURN_NULL(server, avahi_is_valid_service_type_strict(type), AVAHI_ERR_INVALID_SERVICE_TYPE);
    AVAHI_CHECK_VALIDITY_RETURN_NULL(server, AVAHI_FLAGS_VALID(flags, AVAHI_LOOKUP_USE_WIDE_AREA|AVAHI_LOOKUP_USE_MULTICAST|AVAHI_LOOKUP_NO_TXT|AVAHI_LOOKUP_NO_ADDRESS), AVAHI_ERR_INVALID_FLAGS);

    if (!domain)
        domain = server->domain_name;

    if ((ret = avahi_service_name_join(n, sizeof(n), name, type, domain)) < 0) {
        avahi_server_set_errno(server, ret);
        return NULL;
    }

    if (!(r = avahi_new(AvahiSServiceResolver, 1))) {
        avahi_server_set_errno(server, AVAHI_ERR_NO_MEMORY);
        return NULL;
    }

    r->server = server;
    r->service_name = avahi_strdup(name);
    r->service_type = avahi_normalize_name_strdup(type);
    r->domain_name = avahi_normalize_name_strdup(domain);
    r->callback = callback;
    r->userdata = userdata;
    r->address_protocol = aprotocol;
    r->srv_record = r->txt_record = r->address_record = NULL;
    r->srv_flags = r->txt_flags = r->address_flags = 0;
    r->interface = interface;
    r->protocol = protocol;
    r->user_flags = flags;
    r->record_browser_a = r->record_browser_aaaa = r->record_browser_srv = r->record_browser_txt = NULL;
    r->time_event = NULL;
    AVAHI_LLIST_PREPEND(AvahiSServiceResolver, resolver, server->service_resolvers, r);

#ifdef HAVE_BONJOUR
    r->defer_time_event = NULL;
    r->host = NULL;
    r->watch = NULL;
    r->watch_a = NULL;
    r->watch_a6 = NULL;
    r->client = NULL;
    r->client_a = NULL;
    r->client_a6 = NULL;
    avahi_resolve_service_start(server, r);
#else
    k = avahi_key_new(n, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_SRV);
    r->record_browser_srv = avahi_s_record_browser_new(server, interface, protocol, k, flags & ~(AVAHI_LOOKUP_NO_TXT|AVAHI_LOOKUP_NO_ADDRESS), record_browser_callback, r);
    avahi_key_unref(k);

    if (!r->record_browser_srv) {
        avahi_s_service_resolver_free(r);
        return NULL;
    }

    if (!(flags & AVAHI_LOOKUP_NO_TXT)) {
        k = avahi_key_new(n, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_TXT);
        r->record_browser_txt = avahi_s_record_browser_new(server, interface, protocol, k, flags & ~(AVAHI_LOOKUP_NO_TXT|AVAHI_LOOKUP_NO_ADDRESS),  record_browser_callback, r);
        avahi_key_unref(k);

        if (!r->record_browser_txt) {
            avahi_s_service_resolver_free(r);
            return NULL;
        }
    }
#endif

    start_timeout(r);

    return r;
}

void avahi_s_service_resolver_start(AvahiSServiceResolver *r) {
    assert(r);

    if (r->record_browser_srv)
        avahi_s_record_browser_start_query(r->record_browser_srv);
    if (r->record_browser_txt)
        avahi_s_record_browser_start_query(r->record_browser_txt);
    if (r->record_browser_a)
        avahi_s_record_browser_start_query(r->record_browser_a);
    if (r->record_browser_aaaa)
        avahi_s_record_browser_start_query(r->record_browser_aaaa);
}

void avahi_s_service_resolver_free(AvahiSServiceResolver *r) {
    assert(r);

    AVAHI_LLIST_REMOVE(AvahiSServiceResolver, resolver, r->server->service_resolvers, r);

    if (r->time_event)
        avahi_time_event_free(r->time_event);

    if (r->record_browser_srv)
        avahi_s_record_browser_free(r->record_browser_srv);
    if (r->record_browser_txt)
        avahi_s_record_browser_free(r->record_browser_txt);
    if (r->record_browser_a)
        avahi_s_record_browser_free(r->record_browser_a);
    if (r->record_browser_aaaa)
        avahi_s_record_browser_free(r->record_browser_aaaa);

    if (r->srv_record)
        avahi_record_unref(r->srv_record);
    if (r->txt_record)
        avahi_record_unref(r->txt_record);
    if (r->address_record)
        avahi_record_unref(r->address_record);

#ifdef HAVE_BONJOUR
    if (r->defer_time_event) {
        avahi_time_event_free(r->defer_time_event);
        r->defer_time_event = NULL;
    }

    if (r->host)
        avahi_free(r->host);

    if (r->watch)
        r->server->poll_api->watch_free(r->watch);
    if (r->watch_a)
        r->server->poll_api->watch_free(r->watch_a);
    if (r->watch_a6)
        r->server->poll_api->watch_free(r->watch_a6);

    if (r->client)
        DNSServiceRefDeallocate (r->client);
    if (r->client_a)
        DNSServiceRefDeallocate (r->client_a);
    if (r->client_a6)
        DNSServiceRefDeallocate (r->client_a6);
#endif
    avahi_free(r->service_name);
    avahi_free(r->service_type);
    avahi_free(r->domain_name);
    avahi_free(r);
}

AvahiSServiceResolver *avahi_s_service_resolver_new(
    AvahiServer *server,
    AvahiIfIndex interface,
    AvahiProtocol protocol,
    const char *name,
    const char *type,
    const char *domain,
    AvahiProtocol aprotocol,
    AvahiLookupFlags flags,
    AvahiSServiceResolverCallback callback,
    void* userdata) {
        AvahiSServiceResolver *b;

        b = avahi_s_service_resolver_prepare(server, interface, protocol, name, type, domain, aprotocol, flags, callback, userdata);
        avahi_s_service_resolver_start(b);

        return b;
}